# -*- coding: utf-8 -*-

# Author: Daniel Emaasit <daniel.emaasit@gmail.com>
#         (mostly translation, see implementation details)
# License: BSD 3 clause

"""
The :mod:`pmlearn.gaussian_process` module implements Gaussian Process
based regression and classification.
"""

from .gpr import GaussianProcessRegressor
from .gpr import StudentsTProcessRegressor
from .gpr import SparseGaussianProcessRegressor
# from .gpc import GaussianProcessClassifier
# from . import kernels


__all__ = ['GaussianProcessRegressor',
           'StudentsTProcessRegressor',
           'SparseGaussianProcessRegressor']


"""
Multilayer perceptron
"""

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import numpy as np
import pymc3 as pm
import theano

from ..base import BayesianModel, BayesianClassifierMixin

floatX = theano.config.floatX


class MLPClassifier(BayesianModel, BayesianClassifierMixin):
    """ Multilayer perceptron classification built using PyMC3.

    Fit a Multilayer perceptron classification model and estimate
    model parameters using
    MCMC algorithms or Variational Inference algorithms

    Parameters
    ----------


    Examples
    --------


    Reference
    ----------
    http://twiecki.github.io/blog/2016/06/01/bayesian-deep-learning/
    """
    def __init__(self, n_hidden=5):
        self.n_hidden = n_hidden
        self.num_training_samples = None
        self.num_pred = None
        self.total_size = None

        super(MLPClassifier, self).__init__()

    def create_model(self):
        """

        Returns
        -------

        """
        model_input = theano.shared(np.zeros([self.num_training_samples,
                                              self.num_pred]))

        model_output = theano.shared(np.zeros(self.num_training_samples))

        self.shared_vars = {
            'model_input': model_input,
            'model_output': model_output,
        }

        self.total_size = len(model_output.get_value())

        # Initialize random weights between each layer
        init_1 = np.random.randn(self.num_pred, self.n_hidden).astype(floatX)
        init_2 = np.random.randn(self.n_hidden, self.n_hidden).astype(floatX)
        init_out = np.random.randn(self.n_hidden).astype(floatX)

        model = pm.Model()

        with model:
            # Weights from input to hidden layer
            weights_in_1 = pm.Normal('w_in_1', 0, sd=1,
                                     shape=(self.num_pred, self.n_hidden),
                                     testval=init_1)

            # Weights from 1st to 2nd layer
            weights_1_2 = pm.Normal('w_1_2', 0, sd=1,
                                    shape=(self.n_hidden, self.n_hidden),
                                    testval=init_2)

            # Weights from hidden layer to output
            weights_2_out = pm.Normal('w_2_out', 0, sd=1,
                                      shape=(self.n_hidden,),
                                      testval=init_out)

            # Build neural-network using tanh activation function
            act_1 = pm.math.tanh(pm.math.dot(model_input, weights_in_1))
            act_2 = pm.math.tanh(pm.math.dot(act_1, weights_1_2))
            act_out = pm.math.sigmoid(pm.math.dot(act_2, weights_2_out))

            # Binary classification -> Bernoulli likelihood
            y = pm.Bernoulli('y',
                             act_out,
                             observed=model_output,
                             total_size=self.total_size)
        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(MLPClassifier, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(MLPClassifier, self).load(file_prefix,
                                                 load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']



"""Testing for Logistic regression """

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import pytest
import numpy.testing as npt
import pandas.testing as pdt
import shutil
import tempfile

import numpy as np
from pymc3 import summary
from sklearn.model_selection import train_test_split

from pmlearn.exceptions import NotFittedError
from pmlearn.linear_model import HierarchicalLogisticRegression


class TestHierarchicalLogisticRegression(object):
    def setup_method(self):
        def numpy_invlogit(x):
            return 1 / (1 + np.exp(-x))

        self.num_cats = 3
        self.num_pred = 1
        self.num_samples_per_cat = 1000

        self.alphas = np.random.randn(self.num_cats)
        self.betas = np.random.randn(self.num_cats, self.num_pred)
        # TODO: make this more efficient; right now, it's very explicit
        # so I understand it.
        x_a = np.random.randn(self.num_samples_per_cat, self.num_pred)
        y_a = np.random.binomial(1,
                                 numpy_invlogit(self.alphas[0] +
                                                np.sum(self.betas[0] * x_a, 1)
                                                ))
        x_b = np.random.randn(self.num_samples_per_cat, self.num_pred)
        y_b = np.random.binomial(1,
                                 numpy_invlogit(self.alphas[1] +
                                                np.sum(self.betas[1] * x_b, 1)
                                                ))
        x_c = np.random.randn(self.num_samples_per_cat, self.num_pred)
        y_c = np.random.binomial(1,
                                 numpy_invlogit(self.alphas[2] +
                                                np.sum(self.betas[2] * x_c, 1)
                                                ))

        X = np.concatenate([x_a, x_b, x_c])
        y = np.concatenate([y_a, y_b, y_c])
        cats = np.concatenate([
            np.zeros(self.num_samples_per_cat, dtype=np.int),
            np.ones(self.num_samples_per_cat, dtype=np.int),
            2*np.ones(self.num_samples_per_cat, dtype=np.int)
        ])

        self.X_train, self.X_test, self.cat_train, self.cat_test, \
        self.y_train, self.y_test = train_test_split(
            X, cats, y, test_size=0.4
        )

        self.advi_hlr = HierarchicalLogisticRegression()

        self.test_dir = tempfile.mkdtemp()

    def teardown_method(self):
        shutil.rmtree(self.test_dir)


class TestHierarchicalLogisticRegressionFit(TestHierarchicalLogisticRegression):
    def test_advi_fit_returns_correct_model(self):
        # Note: print is here so PyMC3 output won't overwrite the test name
        print('')
        self.advi_hlr.fit(self.X_train, self.y_train, self.cat_train,
                          minibatch_size=500, inference_args={"n": 50000})

        npt.assert_equal(self.num_cats, self.advi_hlr.num_cats)
        npt.assert_equal(self.num_pred, self.advi_hlr.num_pred)

        #TODO: Figure out best way to test
        #np.testing.assert_almost_equal(self.alphas,
        # self.advi_hlr.trace['alphas'].mean(), decimal=1)
        #np.testing.assert_almost_equal(self.betas,
        # self.advi_hlr.trace['betas'].mean(), decimal=1)

        # For now, just check that the estimated parameters
        # have the correct signs
        npt.assert_equal(
            np.sign(self.alphas),
            np.sign(self.advi_hlr.trace['alpha'].mean(axis=0))
        )
        npt.assert_equal(
            np.sign(self.betas),
            np.sign(self.advi_hlr.trace['beta'].mean(axis=0))
        )


class TestHierarchicalLogisticRegressionPredictProba(
    TestHierarchicalLogisticRegression):

    def test_predict_proba_returns_probabilities(self):
        print('')
        self.advi_hlr.fit(self.X_train, self.y_train, self.cat_train,
                          minibatch_size=500, inference_args={"n": 50000})
        probs = self.advi_hlr.predict_proba(self.X_test, self.cat_test)
        npt.assert_equal(probs.shape, self.y_test.shape)

    def test_predict_proba_returns_probabilities_and_std(self):
        print('')
        self.advi_hlr.fit(self.X_train, self.y_train, self.cat_train,
                          minibatch_size=500, inference_args={"n": 50000})
        probs, stds = self.advi_hlr.predict_proba(self.X_test, self.cat_test,
                                                  return_std=True)
        npt.assert_equal(probs.shape, self.y_test.shape)
        npt.assert_equal(stds.shape, self.y_test.shape)

    def test_predict_proba_raises_error_if_not_fit(self):
        with pytest.raises(NotFittedError):
            advi_hlr = HierarchicalLogisticRegression()
            advi_hlr.predict_proba(self.X_train, self.cat_train)


class TestHierarchicalLogisticRegressionPredict(
    TestHierarchicalLogisticRegression):

    def test_predict_returns_predictions(self):
        print('')
        self.advi_hlr.fit(self.X_train, self.y_train, self.cat_train,
                          minibatch_size=500, inference_args={"n": 50000})
        preds = self.advi_hlr.predict(self.X_test, self.cat_test)
        npt.assert_equal(preds.shape, self.y_test.shape)


class TestHierarchicalLogisticRegressionScore(
    TestHierarchicalLogisticRegression):

    def test_score_scores(self):
        print('')
        self.advi_hlr.fit(self.X_train, self.y_train, self.cat_train,
                          minibatch_size=500, inference_args={"n": 50000})
        score = self.advi_hlr.score(self.X_test, self.y_test, self.cat_test)
        naive_score = np.mean(self.y_test)
        npt.assert_array_less(naive_score, score)


class TestHierarchicalLogisticRegressionSaveandLoad(
    TestHierarchicalLogisticRegression):

    def test_save_and_load_work_correctly(self):
        print('')
        self.advi_hlr.fit(self.X_train, self.y_train, self.cat_train,
                          minibatch_size=500, inference_args={"n": 50000})
        probs1 = self.advi_hlr.predict_proba(self.X_test, self.cat_test)
        self.advi_hlr.save(self.test_dir)

        hlr2 = HierarchicalLogisticRegression()

        hlr2.load(self.test_dir)

        npt.assert_equal(self.advi_hlr.num_cats, hlr2.num_cats)
        npt.assert_equal(self.advi_hlr.num_pred, hlr2.num_pred)
        npt.assert_equal(self.advi_hlr.num_training_samples,
                         hlr2.num_training_samples)
        pdt.assert_frame_equal(summary(self.advi_hlr.trace),
                               summary(hlr2.trace))

        probs2 = hlr2.predict_proba(self.X_test, self.cat_test)

        npt.assert_almost_equal(probs2, probs1, decimal=1)



"""Testing for Linear regression """

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import pytest
import numpy.testing as npt
import pandas.testing as pdt
import shutil
import tempfile

import numpy as np
from pymc3 import summary
from sklearn.linear_model import LinearRegression as skLinearRegression
from sklearn.model_selection import train_test_split

from pmlearn.exceptions import NotFittedError
from pmlearn.linear_model import LinearRegression


class TestLinearRegression(object):

    def setup_method(self):
        self.num_pred = 1
        self.alpha = 2
        self.betas = 3
        self.s = 1

        X = np.random.randn(1000, 1)
        noise = self.s * np.random.randn(1000, 1)
        y = self.betas * X + self.alpha + noise
        y = np.squeeze(y)

        self.X_train, self.X_test, self.y_train, self.y_test = \
            train_test_split(X, y, test_size=0.4)

        self.advi_lr = LinearRegression()
        self.nuts_lr = LinearRegression()

        self.test_dir = tempfile.mkdtemp()

    def teardown_method(self):
        shutil.rmtree(self.test_dir)


class TestLinearRegressionFit(TestLinearRegression):
    def test_advi_fit_returns_correct_model(self):
        # Note: print is here so PyMC3 output won't overwrite the test name
        print('')
        self.advi_lr.fit(self.X_train, self.y_train)

        npt.assert_equal(self.num_pred, self.advi_lr.num_pred)

        npt.assert_almost_equal(self.alpha,
                                self.advi_lr.summary['mean']['alpha__0'],
                                decimal=1)
        npt.assert_almost_equal(self.betas,
                                self.advi_lr.summary['mean']['betas__0_0'],
                                decimal=1)
        npt.assert_almost_equal(self.s,
                                self.advi_lr.summary['mean']['s'],
                                decimal=1)

    def test_nuts_fit_returns_correct_model(self):
        # Note: print is here so PyMC3 output won't overwrite the test name
        print('')
        self.nuts_lr.fit(self.X_train, self.y_train, inference_type='nuts',
                         inference_args={'draws': 2000})

        npt.assert_equal(self.num_pred, self.nuts_lr.num_pred)

        npt.assert_almost_equal(self.alpha,
                                self.nuts_lr.summary['mean']['alpha__0'],
                                decimal=1)
        npt.assert_almost_equal(self.betas,
                                self.nuts_lr.summary['mean']['betas__0_0'],
                                decimal=1)
        npt.assert_almost_equal(self.s,
                                self.nuts_lr.summary['mean']['s'],
                                decimal=1)


class TestLinearRegressionPredict(TestLinearRegression):
    def test_predict_returns_predictions(self):
        print('')
        self.advi_lr.fit(self.X_train, self.y_train)
        preds = self.advi_lr.predict(self.X_test)
        npt.assert_equal(preds.shape, self.y_test.shape)

    def test_predict_returns_mean_predictions_and_std(self):
        print('')
        self.advi_lr.fit(self.X_train, self.y_train)
        preds, stds = self.advi_lr.predict(self.X_test, return_std=True)
        npt.assert_equal(preds.shape, self.y_test.shape)
        npt.assert_equal(stds.shape, self.y_test.shape)

    def test_predict_raises_error_if_not_fit(self):
        with pytest.raises(NotFittedError):
            lr = LinearRegression()
            lr.predict(self.X_train)


class TestLinearRegressionScore(TestLinearRegression):
    def test_score_matches_sklearn_performance(self):
        print('')
        sklr = skLinearRegression()
        sklr.fit(self.X_train, self.y_train)
        sklr_score = sklr.score(self.X_test, self.y_test)

        self.advi_lr.fit(self.X_train, self.y_train)
        score = self.advi_lr.score(self.X_test, self.y_test)
        npt.assert_almost_equal(sklr_score, score, decimal=1)


class TestLinearRegressionSaveandLoad(TestLinearRegression):
    def test_save_and_load_work_correctly(self):
        print('')
        self.advi_lr.fit(self.X_train, self.y_train)
        score1 = self.advi_lr.score(self.X_test, self.y_test)
        self.advi_lr.save(self.test_dir)

        lr2 = LinearRegression()
        lr2.load(self.test_dir)

        npt.assert_equal(self.advi_lr.inference_type, lr2.inference_type)
        npt.assert_equal(self.advi_lr.num_pred, lr2.num_pred)
        npt.assert_equal(self.advi_lr.num_training_samples,
                         lr2.num_training_samples)
        pdt.assert_frame_equal(summary(self.advi_lr.trace), summary(lr2.trace))

        score2 = lr2.score(self.X_test, self.y_test)
        npt.assert_almost_equal(score1, score2, decimal=1)



"""Dirichlet Process Mixture Model. """

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import numpy as np
import pymc3 as pm
import theano
import theano.tensor as tt

from ..exceptions import NotFittedError
from ..base import BayesianModel, BayesianDensityMixin
from .util import logp_gmix


class DirichletProcessMixture(BayesianModel, BayesianDensityMixin):
    """
    Custom Dirichlet Process Mixture Model built using PyMC3.
    """

    def __init__(self):
        super(DirichletProcessMixture, self).__init__()
        self.num_truncate = None

    def create_model(self):
        """
        Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will raise
        an error. See http://docs.pymc.io/advanced_theano.html

        The DensityDist class is used as the likelihood term. The second
        argument, logp_gmix(mus, pi, np.eye(D)), is a python function which
        recieves observations (denoted by 'value') and returns the tensor
        representation of the log-likelihood.

        Returns
        ----------
        the PyMC3 model
        """
        model_input = theano.shared(np.zeros([self.num_training_samples,
                                              self.num_pred]))

        # model_output = theano.shared(np.zeros(self.num_training_samples))

        # model_truncate = theano.shared(np.zeros(self.num_training_samples,
        #                                     dtype='int'))

        self.shared_vars = {
            'model_input': model_input
            # ,
            # 'model_output': model_output,
            # 'model_truncate': model_truncate
        }

        # Log likelihood of normal distribution
        # def logp_normal(mu, tau, value):
        #     # log probability of individual samples
        #     k = tau.shape[0]
        #
        #     def delta(mu):
        #         return value - mu
        #     # delta = lambda mu: value - mu
        #     return (-1 / 2.) * (k * T.log(2 * np.pi) + T.log(1./det(tau)) +
        #                          (delta(mu).dot(tau) * delta(
        # mu)).sum(axis=1))

        # Log likelihood of Gaussian mixture distribution
        # def logp_gmix(mus, pi, tau):
        #     def logp_(value):
        #         logps = [T.log(pi[i]) + logp_normal(mu, tau, value)
        #                  for i, mu in enumerate(mus)]
        #
        #         return T.sum(
        # logsumexp(T.stacklists(logps)[:, :self.num_training_samples],
        # axis=0))
        #
        #     return logp_

        def stick_breaking(v):
            portion_remaining = tt.concatenate(
                [[1], tt.extra_ops.cumprod(1 - v)[:-1]])
            return v * portion_remaining

        model = pm.Model()

        with model:

            K = self.num_truncate
            D = self.num_pred

            alpha = pm.Gamma('alpha', 1.0, 1.0)
            v = pm.Beta('v', 1, alpha, shape=K)
            pi_ = stick_breaking(v)
            pi = pm.Deterministic('pi', pi_/pi_.sum())

            means = tt.stack([pm.Uniform('cluster_center_{}'.format(k),
                                        lower=0.,
                                        upper=10.,
                                        shape=D) for k in range(K)])

            lower = tt.stack([pm.LKJCholeskyCov(
                'cluster_variance_{}'.format(k),
                n=D,
                eta=2.,
                sd_dist=pm.HalfNormal.dist(sd=1.)) for k in range(K)])

            chol = tt.stack([pm.expand_packed_triangular(
                D, lower[k]) for k in range(K)])

            component_dists = [pm.MvNormal(
                'component_dist_%d' % k,
                mu=means[k],
                chol=chol[k],
                shape=D) for k in range(K)]

            # rand = [pm.MvNormal(
            # 'rand_{}'.format(k),
            # mu=means[k], chol=Chol[k], shape=D) for k in range(K)]
            rand = pm.Normal.dist(0, 1).random

            X = pm.DensityDist(
                'X',
                logp_gmix(
                    mus=component_dists, pi=pi,
                    tau=np.eye(D),
                    num_training_samples=model_input.get_value().shape[0]),
                observed=model_input, random=rand)

        return model

    def predict_proba(self, X, return_std=False):
        """
        Predicts probabilities of new data with a trained Dirichlet Process
        Mixture Model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        cats : numpy array, shape [n_samples, ]

        return_std : Boolean flag
           Boolean flag of whether to return standard deviations with mean
           probabilities. Defaults to False.
        """

        if self.trace is None:
            raise NotFittedError('Run fit on the model before predict.')

        # num_samples = X.shape[0]

        if self.cached_model is None:
            self.cached_model = self.create_model()

        self._set_shared_vars({'model_input': X})
        _vars = self.cached_model.free_RVs[8:11]

        ppc = pm.sample_ppc(self.trace,
                            model=self.cached_model,
                            vars=_vars,
                            samples=2000,
                            size=len(X))
        return(ppc)

        # if return_std:
        #     return ppc['X'].mean(axis=0), ppc['X'].std(axis=0)
        # else:
        #     return ppc['X'].mean(axis=0)

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            # 'num_cats': self.num_cats,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(DirichletProcessMixture, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(DirichletProcessMixture, self).load(
            file_prefix, load_custom_params=True)

        self.inference_type = params['inference_type']
        # self.num_cats = params['num_cats']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']



"""Testing for Gaussian process regression """

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import pytest
import numpy.testing as npt
import pandas.testing as pdt
import shutil
import tempfile

import numpy as np
import pymc3 as pm
from pymc3 import summary
from sklearn.gaussian_process import \
    GaussianProcessRegressor as skGaussianProcessRegressor
from sklearn.model_selection import train_test_split


from pmlearn.exceptions import NotFittedError
from pmlearn.gaussian_process import (GaussianProcessRegressor,
                                      StudentsTProcessRegressor,
                                      SparseGaussianProcessRegressor)


class TestGaussianProcessRegressor(object):

    def setup_method(self):
        self.num_pred = 1
        self.num_training_samples = 300

        self.length_scale = 1.0
        self.signal_variance = 0.1
        self.noise_variance = 0.1

        X = np.linspace(start=0, stop=10,
                        num=self.num_training_samples)[:, None]

        cov_func = self.signal_variance ** 2 * pm.gp.cov.ExpQuad(
            1, self.length_scale)
        mean_func = pm.gp.mean.Zero()

        f_true = np.random.multivariate_normal(
            mean_func(X).eval(),
            cov_func(X).eval() + 1e-8 * np.eye(self.num_training_samples),
            1).flatten()
        y = f_true + \
            self.noise_variance * np.random.randn(self.num_training_samples)

        self.X_train, self.X_test, self.y_train, self.y_test = \
            train_test_split(X, y, test_size=0.3)

        self.advi_gpr = GaussianProcessRegressor()
        
        self.test_dir = tempfile.mkdtemp()

    def teardown_method(self):
        """Tear down
        """
        shutil.rmtree(self.test_dir)


class TestGaussianProcessRegressorFit(TestGaussianProcessRegressor):
    def test_advi_fit_returns_correct_model(self):
        # This print statement ensures PyMC3 output won't overwrite
        # the test name
        print('')
        self.advi_gpr.fit(self.X_train, self.y_train,
                          inference_args={"n": 25000})

        npt.assert_equal(self.num_pred, self.advi_gpr.num_pred)
        npt.assert_almost_equal(
            self.signal_variance,
            self.advi_gpr.summary['mean']['signal_variance__0'],
            0)
        npt.assert_almost_equal(
            self.length_scale,
            self.advi_gpr.summary['mean']['length_scale__0_0'],
            0)
        npt.assert_almost_equal(
            self.noise_variance,
            self.advi_gpr.summary['mean']['noise_variance__0'],
            0)


class TestGaussianProcessRegressorPredict(TestGaussianProcessRegressor):
    def test_predict_returns_predictions(self):
        print('')
        self.advi_gpr.fit(self.X_train, self.y_train,
                          inference_args={"n": 25000})
        preds = self.advi_gpr.predict(self.X_test)
        npt.assert_equal(self.y_test.shape, preds.shape)

    def test_predict_returns_mean_predictions_and_std(self):
        print('')
        self.advi_gpr.fit(self.X_train, self.y_train,
                          inference_args={"n": 25000})
        preds, stds = self.advi_gpr.predict(self.X_test, return_std=True)
        npt.assert_equal(self.y_test.shape, preds.shape)
        npt.assert_equal(self.y_test.shape, stds.shape)

    def test_predict_raises_error_if_not_fit(self):
        print('')
        with pytest.raises(NotFittedError):
            advi_gpr = GaussianProcessRegressor()
            advi_gpr.predict(self.X_train)


class TestGaussianProcessRegressorScore(TestGaussianProcessRegressor):
    def test_score_matches_sklearn_performance(self):
        print('')
        sk_gpr = skGaussianProcessRegressor()
        sk_gpr.fit(self.X_train, self.y_train)
        sk_gpr_score = sk_gpr.score(self.X_test, self.y_test)

        self.advi_gpr.fit(self.X_train, self.y_train,
                          inference_args={"n": 25000})
        advi_gpr_score = self.advi_gpr.score(self.X_test, self.y_test)

        npt.assert_almost_equal(sk_gpr_score, advi_gpr_score, 1)


class TestGaussianProcessRegressorSaveAndLoad(TestGaussianProcessRegressor):
    def test_save_and_load_work_correctly(self):
        print('')
        self.advi_gpr.fit(self.X_train, self.y_train,
                          inference_args={"n": 25000})
        score1 = self.advi_gpr.score(self.X_test, self.y_test)
        self.advi_gpr.save(self.test_dir)

        gpr2 = GaussianProcessRegressor()
        gpr2.load(self.test_dir)

        npt.assert_equal(self.advi_gpr.inference_type, gpr2.inference_type)
        npt.assert_equal(self.advi_gpr.num_pred, gpr2.num_pred)
        npt.assert_equal(self.advi_gpr.num_training_samples,
                         gpr2.num_training_samples)
        pdt.assert_frame_equal(summary(self.advi_gpr.trace),
                               summary(gpr2.trace))

        score2 = gpr2.score(self.X_test, self.y_test)
        npt.assert_almost_equal(score1, score2, 0)


class TestStudentsTProcessRegressor(object):

    def setup_method(self):
        self.num_pred = 1
        self.num_training_samples = 500

        self.length_scale = 1.0
        self.signal_variance = 0.1
        self.noise_variance = 0.1
        self.degrees_of_freedom = 1.0

        X = np.linspace(start=0, stop=10,
                        num=self.num_training_samples)[:, None]

        cov_func = self.signal_variance ** 2 * pm.gp.cov.ExpQuad(
            1, self.length_scale)
        mean_func = pm.gp.mean.Zero()

        f_true = np.random.multivariate_normal(
            mean_func(X).eval(),
            cov_func(X).eval() + 1e-8 * np.eye(self.num_training_samples),
            1).flatten()
        y = f_true + \
            self.noise_variance * \
            np.random.standard_t(self.degrees_of_freedom,
                                 size=self.num_training_samples)

        self.X_train, self.X_test, self.y_train, self.y_test = \
            train_test_split(X, y, test_size=0.3)

        self.advi_stpr = StudentsTProcessRegressor()

        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)


class TestStudentsTProcessRegressorFit(TestStudentsTProcessRegressor):
    def test_advi_fit_returns_correct_model(self):
        # This print statement ensures PyMC3 output won't overwrite
        # the test name
        print('')
        self.advi_stpr.fit(self.X_train, self.y_train,
                           inference_args={"n": 25000})

        npt.assert_equal(self.num_pred, self.advi_stpr.num_pred)
        npt.assert_almost_equal(
            self.signal_variance,
            self.advi_stpr.summary['mean']['signal_variance__0'],
            0)
        npt.assert_almost_equal(
            self.length_scale,
            self.advi_stpr.summary['mean']['length_scale__0_0'],
            0)
        npt.assert_almost_equal(
            self.noise_variance,
            self.advi_stpr.summary['mean']['noise_variance__0'],
            0)


class TestStudentsTProcessRegressorPredict(TestStudentsTProcessRegressor):
    def test_predict_returns_predictions(self):
        print('')
        self.advi_stpr.fit(self.X_train, self.y_train,
                           inference_args={"n": 25000})
        preds = self.advi_stpr.predict(self.X_test)
        npt.assert_equal(self.y_test.shape, preds.shape)

    def test_predict_returns_mean_predictions_and_std(self):
        print('')
        self.advi_stpr.fit(self.X_train, self.y_train,
                           inference_args={"n": 25000})
        preds, stds = self.advi_stpr.predict(self.X_test, return_std=True)
        npt.assert_equal(self.y_test.shape, preds.shape)
        npt.assert_equal(self.y_test.shape, stds.shape)

    def test_predict_raises_error_if_not_fit(self):
        print('')
        with pytest.raises(NotFittedError):
            advi_stpr = StudentsTProcessRegressor()
            advi_stpr.predict(self.X_train)


class TestStudentsTProcessRegressorScore(TestStudentsTProcessRegressor):
    def test_score_matches_sklearn_performance(self):
        print('')
        sk_gpr = skGaussianProcessRegressor()
        sk_gpr.fit(self.X_train, self.y_train)
        sk_gpr_score = sk_gpr.score(self.X_test, self.y_test)

        self.advi_stpr.fit(self.X_train, self.y_train,
                           inference_args={"n": 25000})
        advi_stpr_score = self.advi_stpr.score(self.X_test, self.y_test)

        npt.assert_almost_equal(sk_gpr_score, advi_stpr_score, 0)


class TestStudentsTProcessRegressorSaveAndLoad(TestStudentsTProcessRegressor):
    def test_save_and_load_work_correctly(self):
        print('')
        self.advi_stpr.fit(self.X_train, self.y_train,
                           inference_args={"n": 25000})
        score1 = self.advi_stpr.score(self.X_test, self.y_test)
        self.advi_stpr.save(self.test_dir)

        stpr2 = StudentsTProcessRegressor()
        stpr2.load(self.test_dir)

        npt.assert_equal(self.advi_stpr.inference_type, stpr2.inference_type)
        npt.assert_equal(self.advi_stpr.num_pred, stpr2.num_pred)
        npt.assert_equal(self.advi_stpr.num_training_samples,
                         stpr2.num_training_samples)
        pdt.assert_frame_equal(summary(self.advi_stpr.trace),
                               summary(stpr2.trace))

        score2 = stpr2.score(self.X_test, self.y_test)
        npt.assert_almost_equal(score1, score2, 0)


class TestSparseGaussianProcessRegressor(object):

    def setup_method(self):
        self.num_pred = 1
        self.num_training_samples = 1000

        self.length_scale = 1.0
        self.signal_variance = 0.1
        self.noise_variance = 0.1

        X = np.linspace(start=0, stop=10,
                        num=self.num_training_samples)[:, None]

        cov_func = self.signal_variance ** 2 * pm.gp.cov.ExpQuad(
            1, self.length_scale)
        mean_func = pm.gp.mean.Zero()

        f_true = np.random.multivariate_normal(
            mean_func(X).eval(),
            cov_func(X).eval() + 1e-8 * np.eye(self.num_training_samples),
            1).flatten()
        y = f_true + \
            self.noise_variance * np.random.randn(self.num_training_samples)

        self.X_train, self.X_test, self.y_train, self.y_test = \
            train_test_split(X, y, test_size=0.3)

        self.advi_sgpr = SparseGaussianProcessRegressor()

        self.test_dir = tempfile.mkdtemp()

    def teardown_method(self):
        """Tear down
        """
        shutil.rmtree(self.test_dir)


class TestSparseGaussianProcessRegressorFit(TestSparseGaussianProcessRegressor):
    def test_advi_fit_returns_correct_model(self):
        # This print statement ensures PyMC3 output won't overwrite
        # the test name
        print('')
        self.advi_sgpr.fit(self.X_train, self.y_train)

        npt.assert_equal(self.num_pred, self.advi_sgpr.num_pred)
        npt.assert_almost_equal(
            self.signal_variance,
            self.advi_sgpr.summary['mean']['signal_variance__0'],
            0)
        npt.assert_almost_equal(
            self.length_scale,
            self.advi_sgpr.summary['mean']['length_scale__0_0'],
            0)
        npt.assert_almost_equal(
            self.noise_variance,
            self.advi_sgpr.summary['mean']['noise_variance__0'],
            0)


class TestSparseGaussianProcessRegressorPredict(
    TestSparseGaussianProcessRegressor):

    def test_predict_returns_predictions(self):
        print('')
        self.advi_sgpr.fit(self.X_train, self.y_train,
                           inference_args={"n": 25000})
        preds = self.advi_sgpr.predict(self.X_test)
        npt.assert_equal(self.y_test.shape, preds.shape)

    def test_predict_returns_mean_predictions_and_std(self):
        print('')
        self.advi_sgpr.fit(self.X_train, self.y_train,
                           inference_args={"n": 25000})
        preds, stds = self.advi_sgpr.predict(self.X_test, return_std=True)
        npt.assert_equal(self.y_test.shape, preds.shape)
        npt.assert_equal(self.y_test.shape, stds.shape)

    def test_predict_raises_error_if_not_fit(self):
        print('')
        with pytest.raises(NotFittedError):
            advi_sgpr = SparseGaussianProcessRegressor()
            advi_sgpr.predict(self.X_train)


class TestSparseGaussianProcessRegressorScore(
    TestSparseGaussianProcessRegressor):

    def test_score_matches_sklearn_performance(self):
        print('')
        sk_gpr = skGaussianProcessRegressor()
        sk_gpr.fit(self.X_train, self.y_train)
        sk_gpr_score = sk_gpr.score(self.X_test, self.y_test)

        self.advi_sgpr.fit(self.X_train, self.y_train)
        advi_sgpr_score = self.advi_sgpr.score(self.X_test, self.y_test)

        npt.assert_almost_equal(sk_gpr_score, advi_sgpr_score, 0)


class TestSparseGaussianProcessRegressorSaveAndLoad(
    TestSparseGaussianProcessRegressor):

    def test_save_and_load_work_correctly(self):
        print('')
        self.advi_sgpr.fit(self.X_train, self.y_train)
        score1 = self.advi_sgpr.score(self.X_test, self.y_test)
        self.advi_sgpr.save(self.test_dir)

        sgpr2 = SparseGaussianProcessRegressor()
        sgpr2.load(self.test_dir)

        npt.assert_equal(self.advi_sgpr.inference_type, sgpr2.inference_type)
        npt.assert_equal(self.advi_sgpr.num_pred, sgpr2.num_pred)
        npt.assert_equal(self.advi_sgpr.num_training_samples,
                         sgpr2.num_training_samples)
        pdt.assert_frame_equal(summary(self.advi_sgpr.trace),
                               summary(sgpr2.trace))

        score2 = sgpr2.score(self.X_test, self.y_test)
        npt.assert_almost_equal(score1, score2, 0)



"""
Generalized Linear models.
"""

# Authors: Nicole Carlson <nicole@parsingscience.com>
#          Daniel <daniel.emaasit@gmail.com>
# License: BSD 3 clause

import numpy as np
import pymc3 as pm
import theano
import theano.tensor as tt
from sklearn.base import ClassifierMixin
from sklearn.metrics import accuracy_score

from ..base import BayesianModel, BayesianRegressorMixin
from ..exceptions import NotFittedError


class BayesianLinearClassifierMixin(ClassifierMixin):
    """Mixin for linear classifiers models in pmlearn

    """
    def fit(self, X, y, cats, inference_type='advi', minibatch_size=None,
            inference_args=None):
        """
        Train the Hierarchical Logistic Regression model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        y : numpy array, shape [n_samples, ]

        cats : numpy array, shape [n_samples, ]

        inference_type : string, specifies which inference method to call.
           Defaults to 'advi'. Currently, only 'advi' and 'nuts' are supported

        minibatch_size : number of samples to include in each minibatch for
           ADVI, defaults to None, so minibatch is not run by default

        inference_args : dict, arguments to be passed to the inference methods.
           Check the PyMC3 docs for permissable values. If no arguments are
           specified, default values will be set.
        """
        self.num_cats = len(np.unique(cats))
        self.num_training_samples, self.num_pred = X.shape

        self.inference_type = inference_type

        if y.ndim != 1:
            y = np.squeeze(y)

        if not inference_args:
            inference_args = self._set_default_inference_args()

        if self.cached_model is None:
            self.cached_model = self.create_model()

        if minibatch_size:
            with self.cached_model:
                minibatches = {
                    self.shared_vars['model_input']: pm.Minibatch(
                        X, batch_size=minibatch_size),
                    self.shared_vars['model_output']: pm.Minibatch(
                        y, batch_size=minibatch_size),
                    self.shared_vars['model_cats']: pm.Minibatch(
                        cats, batch_size=minibatch_size)
                }

                inference_args['more_replacements'] = minibatches
        else:
            self._set_shared_vars({
                'model_input': X,
                'model_output': y,
                'model_cats': cats
            })

        self._inference(inference_type, inference_args)

        return self

    def predict_proba(self, X, cats, return_std=False):
        """
        Predicts probabilities of new data with a trained Hierarchical
        Logistic Regression

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        cats : numpy array, shape [n_samples, ]

        return_std : Boolean flag of whether to return standard deviations with
        mean probabilities. Defaults to False.
        """

        if self.trace is None:
            raise NotFittedError('Run fit on the model before predict.')

        num_samples = X.shape[0]

        if self.cached_model is None:
            self.cached_model = self.create_model()

        self._set_shared_vars({
            'model_input': X,
            'model_output': np.zeros(num_samples, dtype='int'),
            'model_cats': cats
        })

        ppc = pm.sample_ppc(self.trace, model=self.cached_model, samples=2000)

        if return_std:
            return ppc['y'].mean(axis=0), ppc['y'].std(axis=0)
        else:
            return ppc['y'].mean(axis=0)

    def predict(self, X, cats):
        """
        Predicts labels of new data with a trained model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        cats : numpy array, shape [n_samples, ]
        """
        ppc_mean = self.predict_proba(X, cats)

        pred = ppc_mean > 0.5

        return pred

    def score(self, X, y, cats):
        """
        Scores new data with a trained model.

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        y : numpy array, shape [n_samples, ]

        cats : numpy array, shape [n_samples, ]
        """

        return accuracy_score(y, self.predict(X, cats))


class LinearRegression(BayesianModel, BayesianRegressorMixin):
    """
    Linear Regression built using PyMC3.
    """
    def __init__(self):
        super(LinearRegression, self).__init__()

    def create_model(self):
        """
        Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will
        raise an error. See http://docs.pymc.io/advanced_theano.html

        Returns
        ----------
        the PyMC3 model
        """
        model_input = theano.shared(
            np.zeros([self.num_training_samples, self.num_pred]))

        model_output = theano.shared(np.zeros(self.num_training_samples))

        self.shared_vars = {
            'model_input': model_input,
            'model_output': model_output,
        }

        model = pm.Model()

        with model:
            alpha = pm.Normal('alpha', mu=0, sd=100, shape=1)
            betas = pm.Normal('betas', mu=0, sd=100, shape=(1, self.num_pred))

            s = pm.HalfNormal('s', tau=1)

            mean = alpha + tt.sum(betas * model_input, 1)

            y = pm.Normal('y', mu=mean, sd=s, observed=model_output)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(LinearRegression, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(LinearRegression, self).load(file_prefix,
                                                    load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']




"""Unit tests for :mod:`pmlearn.base`. """

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import pytest

from ..base import BayesianModel


class TestBayesianModel(object):
    """Test class for base bayesian model

    """
    def test_create_model_raises_not_implemented_error(self):
        """Assert that NotImplementedError is raised

        """
        with pytest.raises(NotImplementedError):
            bm = BayesianModel()
            bm.create_model()

    # def test_fit_raises_not_implemented_error(self):
    #     with pytest.raises(NotImplementedError):
    #         bm = BayesianModel()
    #         bm.fit()
    #
    # def test_predict_raises_not_implemented_error(self):
    #     with pytest.raises(NotImplementedError):
    #         bm = BayesianModel()
    #         bm.predict()
    #
    # def test_score_raises_not_implemented_error(self):
    #     with pytest.raises(NotImplementedError):
    #         bm = BayesianModel()
    #         bm.score()


"""Testing for Naive Bayes regression """

# Authors: Remi Louf <remilouf@gmail.com>, <remi@sounds.am>
#          Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import pytest
import numpy.testing as npt
import pandas.testing as pdt
import shutil
import tempfile

import numpy as np
import pymc3 as pm
from pymc3 import summary
from sklearn.naive_bayes import GaussianNB as skGaussianNB
from sklearn.model_selection import train_test_split


from pmlearn.exceptions import NotFittedError
from pmlearn.naive_bayes import GaussianNB


class TestGaussianNB(object):

    def setup_method(self):
        self.num_pred = 1
        self.num_training_samples = 300

        self.length_scale = 1.0
        self.signal_variance = 0.1
        self.noise_variance = 0.1

        X = np.linspace(start=0, stop=10,
                        num=self.num_training_samples)[:, None]

        cov_func = self.signal_variance ** 2 * pm.gp.cov.ExpQuad(
            1, self.length_scale)
        mean_func = pm.gp.mean.Zero()

        f_true = np.random.multivariate_normal(
            mean_func(X).eval(),
            cov_func(X).eval() + 1e-8 * np.eye(self.num_training_samples),
            1).flatten()
        y = f_true + \
            self.noise_variance * np.random.randn(self.num_training_samples)

        self.X_train, self.X_test, self.y_train, self.y_test = \
            train_test_split(X, y, test_size=0.3)

        self.advi_gnb = GaussianNB()

        self.test_dir = tempfile.mkdtemp()

    def teardown_method(self):
        """Tear down
        """
        shutil.rmtree(self.test_dir)


class TestGaussianNBFit(TestGaussianNB):
    def test_advi_fit_returns_correct_model(self):
        # This print statement ensures PyMC3 output won't overwrite
        # the test name
        print('')
        self.advi_gnb.fit(self.X_train, self.y_train,
                          inference_args={"n": 25000})

        npt.assert_equal(self.num_pred, self.advi_gnb.num_pred)
        npt.assert_almost_equal(
            self.signal_variance,
            self.advi_gnb.summary['mean']['signal_variance__0'],
            0)
        npt.assert_almost_equal(
            self.length_scale,
            self.advi_gnb.summary['mean']['length_scale__0_0'],
            0)
        npt.assert_almost_equal(
            self.noise_variance,
            self.advi_gnb.summary['mean']['noise_variance__0'],
            0)


# -*- coding: utf-8 -*-

# Author: Daniel Emaasit <daniel.emaasit@gmail.com>
#         (mostly translation, see implementation details)
# License: BSD 3 clause

"""
The :mod:`pmlearn.linear_model` module implements generalized linear models.
It includes Bayesian Regression, Bayesian Logistic Regression, Hierarchical
Logistic Regression.
"""

from .base import LinearRegression
from .logistic import (HierarchicalLogisticRegression,
                       LogisticRegression)


__all__ = ['LinearRegression',
           'HierarchicalLogisticRegression',
           'LogisticRegression']



"""
Probabilistic machine learning module for Python
=================================================
pmlearn is a Python module for practical probabilistic machine
learning built on top of scikit-learn and PymC3.

It aims to provide simple and efficient solutions to learning problems
that are accessible to everybody and reusable in various contexts:
machine-learning as a versatile tool for science and engineering.
See http://pymc-learn.org for complete documentation.
"""

__version__ = '0.0.1.rc2'

__all__ = ['gaussian_process',
           'linear_model',
           'mixture',
           'neural_network',
           'naive_bayes']



"""
Logistic regression models.
"""

# Authors: Nicole Carlson <parsing-science@gmail.com>
#          Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import numpy as np
import pymc3 as pm
import theano
import theano.tensor as tt

from .base import BayesianModel
from .base import BayesianLinearClassifierMixin


class LogisticRegression(BayesianModel, BayesianLinearClassifierMixin):
    """Bayesian Logistic Regression built using PyMC3

    """
    def __init__(self):
        super(LogisticRegression, self).__init__()
        self.num_cats = None

    def create_model(self):
        """
        Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will
        raise an error. See http://docs.pymc.io/advanced_theano.html

        Returns
        ----------
        the PyMC3 model
        """
        model_input = theano.shared(
            np.zeros([self.num_training_samples, self.num_pred]))

        model_output = theano.shared(
            np.zeros(self.num_training_samples, dtype='int'))

        model_cats = theano.shared(
            np.zeros(self.num_training_samples, dtype='int'))

        self.shared_vars = {
            'model_input': model_input,
            'model_output': model_output,
            'model_cats': model_cats
        }

        model = pm.Model()

        with model:
            alpha = pm.Normal('alpha', mu=0, sd=100,
                              shape=(self.num_cats,))
            betas = pm.Normal('beta', mu=0, sd=100,
                              shape=(self.num_cats, self.num_pred))

            c = model_cats

            linear_function = alpha[c] + tt.sum(betas[c] * model_input, 1)

            p = pm.invlogit(linear_function)

            y = pm.Bernoulli('y', p, observed=model_output)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_cats': self.num_cats,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(LogisticRegression, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(LogisticRegression, self).load(
            file_prefix, load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_cats = params['num_cats']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']


class HierarchicalLogisticRegression(BayesianModel,
                                     BayesianLinearClassifierMixin):
    """
    Custom Hierachical Logistic Regression built using PyMC3.
    """

    def __init__(self):
        super(HierarchicalLogisticRegression, self).__init__()
        self.num_cats = None

    def create_model(self):
        """
        Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will
        raise an error. See http://docs.pymc.io/advanced_theano.html

        Returns
        ----------
        the PyMC3 model
        """
        model_input = theano.shared(
            np.zeros([self.num_training_samples, self.num_pred]))

        model_output = theano.shared(
            np.zeros(self.num_training_samples, dtype='int'))

        model_cats = theano.shared(
            np.zeros(self.num_training_samples, dtype='int'))

        self.shared_vars = {
            'model_input': model_input,
            'model_output': model_output,
            'model_cats': model_cats
        }

        model = pm.Model()

        with model:
            mu_alpha = pm.Normal('mu_alpha', mu=0, sd=100)
            sigma_alpha = pm.HalfNormal('sigma_alpha', sd=100)

            mu_beta = pm.Normal('mu_beta', mu=0, sd=100)
            sigma_beta = pm.HalfNormal('sigma_beta', sd=100)

            alpha = pm.Normal('alpha', mu=mu_alpha, sd=sigma_alpha,
                              shape=(self.num_cats,))
            betas = pm.Normal('beta', mu=mu_beta, sd=sigma_beta,
                              shape=(self.num_cats, self.num_pred))

            c = model_cats

            linear_function = alpha[c] + tt.sum(betas[c] * model_input, 1)

            p = pm.invlogit(linear_function)

            y = pm.Bernoulli('y', p, observed=model_output)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_cats': self.num_cats,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(HierarchicalLogisticRegression, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(HierarchicalLogisticRegression,
                       self).load(file_prefix, load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_cats = params['num_cats']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']


"""Base classes for all Bayesian models."""

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#          Nicole Carlson <nicole@parsingscience.com>
# License: BSD 3 clause

import joblib
import numpy as np
import matplotlib.pyplot as plt
import pymc3 as pm
import seaborn as sns
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin, ClassifierMixin, DensityMixin

from .exceptions import NotFittedError


class BayesianModel(BaseEstimator):
    """Base class for all Bayesian models in pymc-learn

    Notes
    -----
    All Bayesian models should specify all the parameters that can be set at
    the class level in their ``__init__`` as explicit keyword arguments
    (no ``*args`` or **kwargs``).
    """
    def __init__(self):
        self.cached_model = None
        self.default_advi_sample_draws = 10000
        self.inference_type = None
        self.num_pred = None
        self.shared_vars = None
        self.summary = None
        self.trace = None
        self.num_training_samples = None

    def create_model(self):
        """Create model
        """
        raise NotImplementedError

    def _set_shared_vars(self, shared_vars):
        """
        Sets theano shared variables for the PyMC3 model.
        """
        for key in shared_vars.keys():
            self.shared_vars[key].set_value(shared_vars[key])

    def _inference(self, inference_type='advi', inference_args=None):
        """
        Calls internal methods for two types of inferences. Raises an error
        if the inference_type is not supported.

        Parameters
        ==========
        inference_type : string, specifies which inference method to call.
           Defaults to 'advi'. Currently, only 'advi' and 'nuts' are supported

        inference_args : dict, arguments to be passed to the inference methods.
           Check the PyMC3 docs to see what is permitted. Defaults to None.
        """
        if inference_type == 'advi':
            self._advi_inference(inference_args)
        elif inference_type == 'nuts':
            self._nuts_inference(inference_args)
        else:
            raise NotFittedError('{} is not a supported type'
                                 ' of inference'.format(inference_type))

    def _advi_inference(self, inference_args):
        """
        Runs variational ADVI and then samples from those results.

        Parameters
        ----------
        inference_args : dict, arguments to be passed to the PyMC3 fit method.
           See PyMC3 doc for permissible values.
        """
        with self.cached_model:
            inference = pm.ADVI()
            approx = pm.fit(method=inference, **inference_args)

        self.approx = approx
        self.trace = approx.sample(draws=self.default_advi_sample_draws)
        self.summary = pm.summary(self.trace)
        self.advi_hist = inference.hist

    def _nuts_inference(self, inference_args):
        """
        Runs NUTS inference.

        Parameters
        ----------
        inference_args : dict, arguments passed to the PyMC3 sample method.
           See PyMC3 doc for permissible values.
        """
        with self.cached_model:
            step = pm.NUTS()
            nuts_trace = pm.sample(step=step, **inference_args)

        self.trace = nuts_trace
        self.summary = pm.summary(self.trace)

    def _set_default_inference_args(self):
        """
        Set default values for inference arguments if none are provided,
        dependent on inference type.

        ADVI
        -----
        callbacks : list containing a parameter stopping check.

        n : number of iterations for ADVI fit, defaults to 200000

        NUTS
        -----
        draws : the number of samples to draw, defaults to 2000
        """
        if self.inference_type == 'advi':
            inference_args = {
                'n': 200000,
                'callbacks': [pm.callbacks.CheckParametersConvergence()]
            }
        elif self.inference_type == 'nuts':
            inference_args = {
                'draws': 2000
            }
        else:
            inference_args = None

        return inference_args

    def save(self, file_prefix, custom_params=None):
        """
        Saves the trace and custom params to files with the given file_prefix.

        Parameters
        ----------
        file_prefix : str, path and prefix used to identify where to save the
        trace for this model.
            Ex: given file_prefix = "path/to/file/"
            This will attempt to save to "path/to/file/trace.pickle"

        custom_params : Dictionary of custom parameters to save.
           Defaults to None
        """
        fileObject = open(file_prefix + 'trace.pickle', 'wb')
        joblib.dump(self.trace, fileObject)
        fileObject.close()

        if custom_params:
            fileObject = open(file_prefix + 'params.pickle', 'wb')
            joblib.dump(custom_params, fileObject)
            fileObject.close()

    def load(self, file_prefix, load_custom_params=False):
        """
        Loads a saved version of the trace, and custom param files with the
        given file_prefix.

        Parameters
        ----------
        file_prefix : str, path and prefix used to identify where to load the
        saved trace for this model.
            Ex: given file_prefix = "path/to/file/"
            This will attempt to load "path/to/file/trace.pickle"

        load_custom_params : Boolean flag to indicate whether custom parameters
        should be loaded. Defaults to False.

        Returns
        ----------
        custom_params : Dictionary of custom parameters
        """
        self.trace = joblib.load(file_prefix + 'trace.pickle')

        custom_params = None
        if load_custom_params:
            custom_params = joblib.load(file_prefix + 'params.pickle')

        return custom_params

    def plot_elbo(self):
        """
        Plot the ELBO values after running ADVI minibatch.
        """
        if self.inference_type != 'advi':
            raise NotFittedError(
                'This method should only be called after calling fit with '
                'ADVI minibatch.'
            )

        sns.set_style("white")
        plt.plot(-self.advi_hist)
        plt.ylabel('ELBO')
        plt.xlabel('iteration')
        sns.despine()


class BayesianRegressorMixin(RegressorMixin):
    """Mixin for regression models in pmlearn

    """
    def fit(self, X, y, inference_type='advi', minibatch_size=None,
            inference_args=None):
        """
        Train the Linear Regression model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        y : numpy array, shape [n_samples, ]

        inference_type : string, specifies which inference method to call.
           Defaults to 'advi'. Currently, only 'advi' and 'nuts' are supported

        minibatch_size : number of samples to include in each minibatch for
           ADVI, defaults to None, so minibatch is not run by default

        inference_args : dict, arguments to be passed to the inference methods.
           Check the PyMC3 docs for permissable values. If no arguments are
           specified, default values will be set.
        """
        self.num_training_samples, self.num_pred = X.shape

        self.inference_type = inference_type

        if y.ndim != 1:
            y = np.squeeze(y)

        if not inference_args:
            inference_args = self._set_default_inference_args()

        if self.cached_model is None:
            self.cached_model = self.create_model()

        if minibatch_size:
            with self.cached_model:
                minibatches = {
                    self.shared_vars['model_input']: pm.Minibatch(
                        X, batch_size=minibatch_size),
                    self.shared_vars['model_output']: pm.Minibatch(
                        y, batch_size=minibatch_size),
                }

                inference_args['more_replacements'] = minibatches
        else:
            self._set_shared_vars({'model_input': X, 'model_output': y})

        self._inference(inference_type, inference_args)

        return self

    def predict(self, X, return_std=False):
        """
        Predicts values of new data with a trained Linear Regression model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        return_std : Boolean flag
           Boolean flag of whether to return standard deviations with mean
           values. Defaults to False.
        """

        if self.trace is None:
            raise NotFittedError('Run fit on the model before predict.')

        num_samples = X.shape[0]

        if self.cached_model is None:
            self.cached_model = self.create_model()

        self._set_shared_vars({'model_input': X,
                               'model_output': np.zeros(num_samples)})

        ppc = pm.sample_ppc(self.trace, model=self.cached_model, samples=2000)

        if return_std:
            return ppc['y'].mean(axis=0), ppc['y'].std(axis=0)
        else:
            return ppc['y'].mean(axis=0)


class BayesianClassifierMixin(ClassifierMixin):
    """Mixin for regression models in pmlearn

    """
    def fit(self, X, y, inference_type='advi', minibatch_size=None,
            inference_args=None):
        """ Train the Multilayer perceptron model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        y : numpy array, shape [n_samples, ]

        inference_type : string, specifies which inference method to call.
        Defaults to 'advi'. Currently, only 'advi' and 'nuts' are supported

        minibatch_size : number of samples to include in each minibatch
        for ADVI, defaults to None, so minibatch is not run by default

        inference_args : dict, arguments to be passed to the inference methods.
        Check the PyMC3 docs for permissable values. If no arguments are
        specified, default values will be set.
        """
        self.num_training_samples, self.num_pred = X.shape

        self.inference_type = inference_type

        if y.ndim != 1:
            y = np.squeeze(y)

        if not inference_args:
            inference_args = self._set_default_inference_args()

        if self.cached_model is None:
            self.cached_model = self.create_model()

        if minibatch_size:
            with self.cached_model:
                minibatches = {
                    self.shared_vars['model_input']: pm.Minibatch(
                        X, batch_size=minibatch_size),
                    self.shared_vars['model_output']: pm.Minibatch(
                        y, batch_size=minibatch_size),
                }

                inference_args['more_replacements'] = minibatches
        else:
            self._set_shared_vars({'model_input': X, 'model_output': y})

        self._inference(inference_type, inference_args)

        return self

    def predict_proba(self, X, return_std=False):
        """ Perform Prediction

        Predicts values of new data with a trained Gaussian Process
        Regression model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        return_std : Boolean
            Whether to return standard deviations with mean values.
            Defaults to False.
        """

        if self.trace is None:
            raise NotFittedError('Run fit on the model before predict.')

        num_samples = X.shape[0]

        if self.cached_model is None:
            self.cached_model = self.create_model()

        self._set_shared_vars({'model_input': X,
                               'model_output': np.zeros(num_samples)})

        ppc = pm.sample_ppc(self.trace, model=self.cached_model, samples=2000)

        if return_std:
            return ppc['y'].mean(axis=0), ppc['y'].std(axis=0)
        else:
            return ppc['y'].mean(axis=0)

    def predict(self, X):
        """
        Predicts labels of new data with a trained model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        """
        ppc_mean = self.predict_proba(X)

        pred = ppc_mean > 0.5

        return pred


class BayesianDensityMixin(DensityMixin):
    """Mixin for regression models in pmlearn

    """
    def fit(self, X, num_components, inference_type='advi',
            minibatch_size=None, inference_args=None):
        """
        Train the Gaussian Mixture Model model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        n_truncate : numpy array, shape [n_samples, ]

        inference_type : string, specifies which inference method to call.
        Defaults to 'advi'. Currently, only 'advi' and 'nuts' are supported

        minibatch_size : number of samples to include in each minibatch for
        ADVI,
        defaults to None, so minibatch is not run by default

        inference_args : dict, arguments to be passed to the inference methods.
        Check the PyMC3 docs for permissable values. If no arguments are
        specified,
        default values will be set.
        """
        self.num_components = num_components
        self.num_training_samples, self.num_pred = X.shape

        self.inference_type = inference_type

        # if y.ndim != 1:
        #     y = np.squeeze(y)

        if not inference_args:
            inference_args = self._set_default_inference_args()

        if self.cached_model is None:
            self.cached_model = self.create_model()

        if minibatch_size:
            with self.cached_model:
                minibatches = {
                    self.shared_vars['model_input']: pm.Minibatch(
                        X, batch_size=minibatch_size)
                    # ,
                    # self.shared_vars['model_output']: pm.Minibatch(
                    # y, batch_size=minibatch_size),
                    # self.shared_vars['model_components']: pm.Minibatch(
                    # components, batch_size=minibatch_size)
                }

                inference_args['more_replacements'] = minibatches
        else:
            self._set_shared_vars({'model_input': X})

        self._inference(inference_type, inference_args)

        return self

    def predict_proba(self, X, return_std=False):
        """
        Predicts probabilities of new data with a trained GaussianMixture Model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        cats : numpy array, shape [n_samples, ]

        return_std : Boolean flag of whether to return standard deviations with
        mean probabilities. Defaults to False.
        """

        if self.trace is None:
            raise NotFittedError('Run fit on the model before predict.')

        # num_samples = X.shape[0]

        if self.cached_model is None:
            self.cached_model = self.create_model()

        self._set_shared_vars({'model_input': X})
        K = self.num_components

        with self.cached_model:
            pi = pm.Dirichlet("probability",
                              a=np.array([1.0, 1.0, 1.0]),
                              shape=K)
            _vars = [pi]

            ppc = pm.sample_ppc(self.trace,
                                # model=self.cached_model,
                                vars=_vars,
                                samples=2000,
                                size=len(X))

        if return_std:
            return ppc['probability'].mean(axis=0), \
                   ppc['probability'].std(axis=0)
        else:
            return ppc['probability'].mean(axis=0)

    def predict(self, X):
        """
        Predicts labels of new data with a trained model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        cats : numpy array, shape [n_samples, ]
        """
        ppc_mean = self.predict_proba(X)

        # pred = ppc_mean > 0.5
        #
        # return pred
        return ppc_mean

    def score(self, X, y, cats):
        """
        Scores new data with a trained model.

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        y : numpy array, shape [n_samples, ]

        cats : numpy array, shape [n_samples, ]
        """
        from sklearn.metrics import accuracy_score
        return accuracy_score(y, self.predict(X, cats))



import unittest
import shutil
import tempfile

import numpy as np
# import pandas as pd
# import pymc3 as pm
# from pymc3 import summary
# from sklearn.mixture import BayesianGaussianMixture as skBayesianGaussianMixture
from sklearn.model_selection import train_test_split

from pmlearn.exceptions import NotFittedError
from pmlearn.mixture import DirichletProcessMixture


class DirichletProcessMixtureTestCase(unittest.TestCase):

    def setUp(self):
        self.num_truncate = 3
        self.num_components = 3
        self.num_pred = 1
        self.num_training_samples = 100

        self.pi = np.array([0.35, 0.4, 0.25])
        self.means = np.array([0, 5, 10])
        self.sigmas = np.array([0.5, 0.5, 1.0])

        self.components = np.random.randint(0,
                                            self.num_components,
                                            self.num_training_samples)

        X = np.random.normal(loc=self.means[self.components],
                             scale=self.sigmas[self.components])
        X.shape = (self.num_training_samples, 1)

        self.X_train, self.X_test = train_test_split(X, test_size=0.3)

        self.test_DPMM = DirichletProcessMixture()
        self.test_nuts_DPMM = DirichletProcessMixture()
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)


# class DirichletProcessMixtureFitTestCase(DirichletProcessMixtureTestCase):
#     def test_advi_fit_returns_correct_model(self):
#         # This print statement ensures PyMC3 output won't overwrite the test name
#         print('')
#         self.test_DPMM.fit(self.X_train)
#
#         self.assertEqual(self.num_pred, self.test_DPMM.num_pred)
#         self.assertEqual(self.num_components, self.test_DPMM.num_components)
#         self.assertEqual(self.num_truncate, self.test_DPMM.num_truncate)
#
#         self.assertAlmostEqual(self.pi[0],
#                                self.test_DPMM.summary['mean']['pi__0'],
#                                0)
#         self.assertAlmostEqual(self.pi[1],
#                                self.test_DPMM.summary['mean']['pi__1'],
#                                0)
#         self.assertAlmostEqual(self.pi[2],
#                                self.test_DPMM.summary['mean']['pi__2'],
#                                0)
#
#         self.assertAlmostEqual(
#             self.means[0],
#             self.test_DPMM.summary['mean']['cluster_center_0__0'],
#             0)
#         self.assertAlmostEqual(
#             self.means[1],
#             self.test_DPMM.summary['mean']['cluster_center_1__0'],
#             0)
#         self.assertAlmostEqual(
#             self.means[2],
#             self.test_DPMM.summary['mean']['cluster_center_2__0'],
#             0)
#
#         self.assertAlmostEqual(
#             self.sigmas[0],
#             self.test_DPMM.summary['mean']['cluster_variance_0__0'],
#             0)
#         self.assertAlmostEqual(
#             self.sigmas[1],
#             self.test_DPMM.summary['mean']['cluster_variance_1__0'],
#             0)
#         self.assertAlmostEqual(
#             self.sigmas[2],
#             self.test_DPMM.summary['mean']['cluster_variance_2__0'],
#             0)
#
#     def test_nuts_fit_returns_correct_model(self):
#         # This print statement ensures PyMC3 output won't overwrite the test name
#         print('')
#         self.test_nuts_DPMM.fit(self.X_train,
#                                 inference_type='nuts',
#                                 inference_args={'draws': 1000,
#                                                 'chains': 2})
#
#         self.assertEqual(self.num_pred, self.test_nuts_DPMM.num_pred)
#         self.assertEqual(self.num_components, self.test_nuts_DPMM.num_components)
#         self.assertEqual(self.num_components, self.test_nuts_DPMM.num_truncate)
#
#         self.assertAlmostEqual(self.pi[0],
#                                self.test_nuts_DPMM.summary['mean']['pi__0'],
#                                0)
#         self.assertAlmostEqual(self.pi[1],
#                                self.test_nuts_DPMM.summary['mean']['pi__1'],
#                                0)
#         self.assertAlmostEqual(self.pi[2],
#                                self.test_nuts_DPMM.summary['mean']['pi__2'],
#                                0)
#
#         self.assertAlmostEqual(
#             self.means[0],
#             self.test_nuts_DPMM.summary['mean']['cluster_center_0__0'],
#             0)
#         self.assertAlmostEqual(
#             self.means[1],
#             self.test_nuts_DPMM.summary['mean']['cluster_center_1__0'],
#             0)
#         self.assertAlmostEqual(
#             self.means[2],
#             self.test_nuts_DPMM.summary['mean']['cluster_center_2__0'],
#             0)
#
#         self.assertAlmostEqual(
#             self.sigmas[0],
#             self.test_nuts_DPMM.summary['mean']['cluster_variance_0__0'],
#             0)
#         self.assertAlmostEqual(
#             self.sigmas[1],
#             self.test_nuts_DPMM.summary['mean']['cluster_variance_1__0'],
#             0)
#         self.assertAlmostEqual(
#             self.sigmas[2],
#             self.test_nuts_DPMM.summary['mean']['cluster_variance_2__0'],
#             0)
#
#
class DirichletProcessMixturePredictTestCase(DirichletProcessMixtureTestCase):
    # def test_predict_returns_predictions(self):
    #     print('')
    #     self.test_DPMM.fit(self.X_train, self.y_train)
    #     preds = self.test_DPMM.predict(self.X_test)
    #     self.assertEqual(self.y_test.shape, preds.shape)

    # def test_predict_returns_mean_predictions_and_std(self):
    #     print('')
    #     self.test_DPMM.fit(self.X_train, self.y_train)
    #     preds, stds = self.test_DPMM.predict(self.X_test, return_std=True)
    #     self.assertEqual(self.y_test.shape, preds.shape)
    #     self.assertEqual(self.y_test.shape, stds.shape)

    def test_predict_raises_error_if_not_fit(self):
        print('')
        with self.assertRaises(NotFittedError) as no_fit_error:
            test_DPMM = DirichletProcessMixture()
            test_DPMM.predict(self.X_train)

        expected = 'Run fit on the model before predict.'
        self.assertEqual(str(no_fit_error.exception), expected)


# class DirichletProcessMixtureScoreTestCase(DirichletProcessMixtureTestCase):
#     def test_score_matches_sklearn_performance(self):
#         print('')
#         skDPMM = skBayesianGaussianMixture(n_components=3)
#         skDPMM.fit(self.X_train)
#         skDPMM_score = skDPMM.score(self.X_test)
#
#         self.test_DPMM.fit(self.X_train)
#         test_DPMM_score = self.test_DPMM.score(self.X_test)
#
#         self.assertAlmostEqual(skDPMM_score, test_DPMM_score, 0)
#
#
# class DirichletProcessMixtureSaveAndLoadTestCase(DirichletProcessMixtureTestCase):
#     def test_save_and_load_work_correctly(self):
#         print('')
#         self.test_DPMM.fit(self.X_train)
#         score1 = self.test_DPMM.score(self.X_test)
#         self.test_DPMM.save(self.test_dir)
#
#         DPMM2 = DirichletProcessMixture()
#         DPMM2.load(self.test_dir)
#
#         self.assertEqual(self.test_DPMM.inference_type, DPMM2.inference_type)
#         self.assertEqual(self.test_DPMM.num_pred, DPMM2.num_pred)
#         self.assertEqual(self.test_DPMM.num_training_samples,
#                          DPMM2.num_training_samples)
#         self.assertEqual(self.test_DPMM.num_truncate, DPMM2.num_truncate)
#
#         pd.testing.assert_frame_equal(summary(self.test_DPMM.trace),
#                                       summary(DPMM2.trace))
#
#         score2 = DPMM2.score(self.X_test)
#         self.assertAlmostEqual(score1, score2, 0)



# -*- coding: utf-8 -*-

# Author: Daniel Emaasit <daniel.emaasit@gmail.com>
#         (mostly translation, see implementation details)
# License: BSD 3 clause

"""
The :mod:`pmlearn.naive_bayes` module implements Naive Bayes algorithms. These
are supervised learning methods based on applying Bayes' theorem with strong
(naive) feature independence assumptions.
"""

from .naive_bayes import GaussianNB


__all__ = ['GaussianNB']



"""Gaussian Mixture Model. """

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import numpy as np
import pymc3 as pm
import theano
import theano.tensor as tt

from ..base import BayesianModel, BayesianDensityMixin


class GaussianMixture(BayesianModel, BayesianDensityMixin):
    """
    Custom Gaussian Mixture Model built using PyMC3.
    """

    def __init__(self):
        super(GaussianMixture, self).__init__()
        self.num_components = None

    def create_model(self):
        """
        Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will raise
        an error. See http://docs.pymc.io/advanced_theano.html

        Returns
        ----------
        the PyMC3 model
        """
        model_input = theano.shared(np.zeros([self.num_training_samples,
                                              self.num_pred]))

        # model_output = theano.shared(np.zeros(self.num_training_samples))

        model_components = theano.shared(np.zeros(self.num_training_samples,
                                                  dtype='int'))

        self.shared_vars = {
            'model_input': model_input
            # ,
            # 'model_output': model_output,
            # 'model_components': model_components
        }

        model = pm.Model()

        with model:

            K = self.num_components
            D = self.num_pred

            pi = pm.Dirichlet("pi", a=np.ones(K) / K, shape=K)
            means = tt.stack([pm.Uniform('cluster_center_{}'.format(k),
                                         lower=0.,
                                         upper=10.,
                                         shape=D) for k in range(K)])
            lower = tt.stack([pm.LKJCholeskyCov(
                'cluster_variance_{}'.format(k),
                n=D,
                eta=2.,
                sd_dist=pm.HalfNormal.dist(sd=1.)) for k in range(K)])
            chol = tt.stack([pm.expand_packed_triangular(
                D, lower[k]) for k in range(K)])

            component_dists = [pm.MvNormal.dist(mu=means[k],
                                                chol=chol[k],
                                                shape=D) for k in range(K)]

            X = pm.Mixture("X", w=pi, comp_dists=component_dists,
                           observed=model_input)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            # 'num_cats': self.num_cats,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(GaussianMixture, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(GaussianMixture, self).load(
            file_prefix, load_custom_params=True)

        self.inference_type = params['inference_type']
        # self.num_cats = params['num_cats']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']



# -*- coding: utf-8 -*-

# Author: Daniel Emaasit <daniel.emaasit@gmail.com>
#         (mostly translation, see implementation details)
# License: BSD 3 clause

"""
The :mod:`pmlearn.mixture` module implements mixture models.
"""

from .gaussian_mixture import GaussianMixture
from .dirichlet_process import DirichletProcessMixture


__all__ = ['GaussianMixture',
           'DirichletProcessMixture']


"""Gaussian process regression. """

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import numpy as np
import pymc3 as pm
import theano

from ..exceptions import NotFittedError
from ..base import BayesianModel, BayesianRegressorMixin

from .kernels import RBF


class GaussianProcessRegressorMixin(BayesianRegressorMixin):
    """Mixin class for Gaussian Process Regression

    """
    def predict(self, X, return_std=False):
        """ Perform Prediction

        Predicts values of new data with a trained Gaussian Process
        Regression model

        Parameters
        ----------
        X : numpy array, shape [n_samples, n_features]

        return_std : Boolean
            Whether to return standard deviations with mean values.
            Defaults to False.
        """

        if self.trace is None:
            raise NotFittedError('Run fit on the model before predict.')

        num_samples = X.shape[0]

        if self.cached_model is None:
            self.cached_model = self.create_model()

        self._set_shared_vars({'model_input': X,
                               'model_output': np.zeros(num_samples)})

        with self.cached_model:
            f_pred = self.gp.conditional("f_pred", X)
            self.ppc = pm.sample_ppc(self.trace,
                                     vars=[f_pred],
                                     samples=2000)

        if return_std:
            return self.ppc['f_pred'].mean(axis=0), \
                   self.ppc['f_pred'].std(axis=0)
        else:
            return self.ppc['f_pred'].mean(axis=0)


class GaussianProcessRegressor(BayesianModel,
                               GaussianProcessRegressorMixin):
    """ Gaussian Process Regression built using PyMC3.

    Fit a Gaussian process model and estimate model parameters using
    MCMC algorithms or Variational Inference algorithms

    Parameters
    ----------
    prior_mean : mean object
        The mean specifying the mean function of the GP. If None is passed,
        the mean "pm.gp.mean.Zero()" is used as default.

    kernel : covariance function (kernel)
        The function specifying the covariance of the GP. If None is passed,
        the kernel "RBF()" is used as default.

    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from pmlearn.gaussian_process import GaussianProcessRegressor
    >>> from pmlearn.gaussian_process.kernels import DotProduct, WhiteKernel
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = DotProduct() + WhiteKernel()
    >>> gpr = GaussianProcessRegressor(kernel=kernel).fit(X, y)
    >>> gpr.score(X, y) # doctest: +ELLIPSIS
    0.3680...
    >>> gpr.predict(X[:2,:], return_std=True) # doctest: +ELLIPSIS
    (array([653.0..., 592.1...]), array([316.6..., 316.6...]))


    Reference
    ----------
    Rasmussen and Williams (2006). Gaussian Processes for Machine Learning.
    """

    def __init__(self, prior_mean=None, kernel=None):
        self.ppc = None
        self.gp = None
        self.num_training_samples = None
        self.num_pred = None
        self.prior_mean = prior_mean
        self.kernel = kernel

        super(GaussianProcessRegressor, self).__init__()

    def create_model(self):
        """ Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will
        raise an error. See http://docs.pymc.io/advanced_theano.html

        Returns
        ----------
        model: the PyMC3 model.
        """
        model_input = theano.shared(np.zeros([self.num_training_samples,
                                              self.num_pred]))

        model_output = theano.shared(np.zeros(self.num_training_samples))

        self.shared_vars = {
            'model_input': model_input,
            'model_output': model_output,
        }

        model = pm.Model()

        with model:
            length_scale = pm.Gamma('length_scale', alpha=2, beta=1,
                                    shape=(1, self.num_pred))
            signal_variance = pm.HalfCauchy('signal_variance', beta=5,
                                            shape=1)
            noise_variance = pm.HalfCauchy('noise_variance', beta=5,
                                           shape=1)

            if self.kernel is None:
                cov_function = signal_variance ** 2 * RBF(
                    input_dim=self.num_pred,
                    ls=length_scale)
            else:
                cov_function = self.kernel

            if self.prior_mean is None:
                mean_function = pm.gp.mean.Zero()
            else:
                mean_function = self.prior_mean

            self.gp = pm.gp.Latent(mean_func=mean_function,
                                   cov_func=cov_function)

            f = self.gp.prior('f', X=model_input.get_value())

            y = pm.Normal('y', mu=f, sd=noise_variance, observed=model_output)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(GaussianProcessRegressor, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(GaussianProcessRegressor, self).load(
            file_prefix, load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']


class StudentsTProcessRegressor(BayesianModel,
                                GaussianProcessRegressorMixin):
    """ StudentsT Process Regression built using PyMC3.

    Fit a StudentsT process model and estimate model parameters using
    MCMC algorithms or Variational Inference algorithms

    Parameters
    ----------
    prior_mean : mean object
        The mean specifying the mean function of the StudentsT process.
        If None is passed, the mean "pm.gp.mean.Zero()" is used as default.

    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from pmlearn.gaussian_process import StudentsTProcessRegressor
    >>> from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = DotProduct() + WhiteKernel()
    >>> spr = StudentsTProcessRegressor(kernel=kernel).fit(X, y)
    >>> spr.score(X, y) # doctest: +ELLIPSIS
    0.3680...
    >>> spr.predict(X[:2,:], return_std=True) # doctest: +ELLIPSIS
    (array([653.0..., 592.1...]), array([316.6..., 316.6...]))


    Reference
    ----------
    Rasmussen and Williams (2006). Gaussian Processes for Machine Learning.
    """

    def __init__(self, prior_mean=None, kernel=None):
        self.ppc = None
        self.gp = None
        self.num_training_samples = None
        self.num_pred = None
        self.prior_mean = prior_mean
        self.kernel = kernel

        super(StudentsTProcessRegressor, self).__init__()

    def create_model(self):
        """ Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will raise
        an error. See http://docs.pymc.io/advanced_theano.html

        Returns
        ----------
        model : the PyMC3 model
        """
        model_input = theano.shared(np.zeros([self.num_training_samples,
                                              self.num_pred]))

        model_output = theano.shared(np.zeros(self.num_training_samples))

        self.shared_vars = {
            'model_input': model_input,
            'model_output': model_output,
        }

        self.gp = None
        model = pm.Model()

        with model:
            length_scale = pm.Gamma('length_scale', alpha=2, beta=0.5,
                                    shape=(1, self.num_pred))
            signal_variance = pm.HalfCauchy('signal_variance', beta=2,
                                            shape=1)
            noise_variance = pm.HalfCauchy('noise_variance', beta=2,
                                           shape=1)
            degrees_of_freedom = pm.Gamma('degrees_of_freedom', alpha=2,
                                          beta=0.1, shape=1)

            if self.kernel is None:
                cov_function = signal_variance ** 2 * RBF(
                    input_dim=self.num_pred,
                    ls=length_scale)
            else:
                cov_function = self.kernel

            if self.prior_mean is None:
                mean_function = pm.gp.mean.Zero()
            else:
                mean_function = self.prior_mean

            self.gp = pm.gp.Latent(mean_func=mean_function,
                                   cov_func=cov_function)

            f = self.gp.prior('f', X=model_input.get_value())

            y = pm.StudentT('y', mu=f, lam=1 / signal_variance,
                            nu=degrees_of_freedom, observed=model_output)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(StudentsTProcessRegressor, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(StudentsTProcessRegressor, self).load(
            file_prefix, load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']


class SparseGaussianProcessRegressor(BayesianModel,
                                     GaussianProcessRegressorMixin):
    """ Sparse Gaussian Process Regression built using PyMC3.

    Fit a Sparse Gaussian process model and estimate model parameters using
    MCMC algorithms or Variational Inference algorithms

    Parameters
    ----------
    prior_mean : mean object
        The mean specifying the mean function of the GP. If None is passed,
        the mean "pm.gp.mean.Zero()" is used as default.

    Examples
    --------
    >>> from sklearn.datasets import make_friedman2
    >>> from pmlearn.gaussian_process import SparseGaussianProcessRegressor
    >>> from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
    >>> X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    >>> kernel = DotProduct() + WhiteKernel()
    >>> sgpr = SparseGaussianProcessRegressor(kernel=kernel).fit(X, y)
    >>> sgpr.score(X, y) # doctest: +ELLIPSIS
    0.3680...
    >>> sgpr.predict(X[:2,:], return_std=True) # doctest: +ELLIPSIS
    (array([653.0..., 592.1...]), array([316.6..., 316.6...]))


    Reference
    ----------
    Rasmussen and Williams (2006). Gaussian Processes for Machine Learning.
    """

    def __init__(self, prior_mean=None, kernel=None):
        self.ppc = None
        self.gp = None
        self.num_training_samples = None
        self.num_pred = None
        self.prior_mean = prior_mean
        self.kernel = kernel

        super(SparseGaussianProcessRegressor, self).__init__()

    def create_model(self):
        """ Creates and returns the PyMC3 model.

        Note: The size of the shared variables must match the size of the
        training data. Otherwise, setting the shared variables later will
        raise an error. See http://docs.pymc.io/advanced_theano.html

        Returns
        ----------
        model : the PyMC3 model
        """
        model_input = theano.shared(np.zeros([self.num_training_samples,
                                              self.num_pred]))

        model_output = theano.shared(np.zeros(self.num_training_samples))

        self.shared_vars = {
            'model_input': model_input,
            'model_output': model_output,
        }

        self.gp = None
        model = pm.Model()

        with model:
            length_scale = pm.Gamma('length_scale', alpha=2, beta=1,
                                    shape=(1, self.num_pred))
            signal_variance = pm.HalfCauchy('signal_variance', beta=5,
                                            shape=1)
            noise_variance = pm.HalfCauchy('noise_variance', beta=5,
                                           shape=1)

            if self.kernel is None:
                cov_function = signal_variance ** 2 * RBF(
                    input_dim=self.num_pred,
                    ls=length_scale)
            else:
                cov_function = self.kernel

            if self.prior_mean is None:
                mean_function = pm.gp.mean.Zero()
            else:
                mean_function = self.prior_mean

            self.gp = pm.gp.MarginalSparse(mean_func=mean_function,
                                           cov_func=cov_function,
                                           approx="FITC")

            # initialize 20 inducing points with K-means
            # gp.util
            Xu = pm.gp.util.kmeans_inducing_points(20,
                                                   X=model_input.get_value())

            y = self.gp.marginal_likelihood('y',
                                            X=model_input.get_value(),
                                            Xu=Xu,
                                            y=model_output.get_value(),
                                            sigma=noise_variance)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }

        super(SparseGaussianProcessRegressor, self).save(file_prefix, params)

    def load(self, file_prefix):
        params = super(SparseGaussianProcessRegressor, self).load(
            file_prefix, load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']



"""
The :mod:`pmlearn.exceptions` module includes all custom warnings and error
classes used across pymc-learn.
"""

__all__ = ['NotFittedError']


class NotFittedError(ValueError, AttributeError):
    """Exception class to raise if estimator is used before fitting.
    This class inherits from both ValueError and AttributeError to help with
    exception handling and backward compatibility.
    Examples
    --------
    >>> from pmlearn.gaussian_process import GaussianProcessRegressor
    >>> from pmlearn.exceptions import NotFittedError
    >>> try:
    ...     GaussianProcessRegressor().predict([[1, 2], [2, 3], [3, 4]])
    ... except NotFittedError as e:
    ...     print(repr(e))
    ...                        # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    NotFittedError('This GaussianProcessRegressor instance is not fitted yet'.)
    """



import numpy as np
# import pymc3 as pm
from pymc3.math import logsumexp
# import theano
import theano.tensor as T
from theano.tensor.nlinalg import det


class logp_gmix(object):
    def __init__(self, mus, pi, tau, num_training_samples):
        self.mus = mus
        self.pi = pi
        self.tau = tau
        self.num_training_samples = num_training_samples

    def __call__(self, value):

        def logp_normal(mu, tau, value):
            # log probability of individual samples
            k = tau.shape[0]

            def delta(mu):
                return value - mu
            # delta = lambda mu: value - mu
            return (-1 / 2.) * (k * T.log(2 * np.pi) + T.log(1. / det(tau)) +
                                (delta(mu).dot(tau) * delta(mu)).sum(axis=1))

        logps = [T.log(self.pi[i]) + logp_normal(mu, self.tau, value)
                 for i, mu in enumerate(self.mus)]

        return T.sum(logsumexp(T.stacklists(logps)[:, :self.num_training_samples],
                               axis=0))

        # Log likelihood of normal distribution
        # def logp_normal(mu, tau, value):
        #     # log probability of individual samples
        #     k = tau.shape[0]
        #
        #     def delta(mu):
        #         return value - mu
        #     # delta = lambda mu: value - mu
        #     return (-1 / 2.) * (k * T.log(2 * np.pi) + T.log(1./det(tau)) +
        #                          (delta(mu).dot(tau) * delta(mu)).sum(axis=1))

        # Log likelihood of Gaussian mixture distribution
        # def logp_gmix(mus, pi, tau):
        #     def logp_(value):
        #         logps = [T.log(pi[i]) + logp_normal(mu, tau, value)
        #                  for i, mu in enumerate(mus)]
        #
        #         return T.sum(logsumexp(T.stacklists(logps)[:, :self.num_training_samples], axis=0))
        #
        #     return logp_



# -*- coding: utf-8 -*-

# Author: Daniel Emaasit <daniel.emaasit@gmail.com>
#         (mostly translation, see implementation details)
# License: BSD 3 clause

"""
The :mod:`pmlearn.neural_network` module includes models based on neural
networks.
"""

from .multilayer_perceptron import MLPClassifier

__all__ = ['MLPClassifier']



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# from builtins import *

from codecs import open
from os.path import realpath, dirname, join
from setuptools import setup, find_packages
import sys
import re

DISTNAME = 'pymc-learn'
DESCRIPTION = "Practical Probabilistic Machine Learning in Python"
AUTHOR = 'Pymc-Learn Team'
AUTHOR_EMAIL = 'daniel.emaasit@gmail.com'
URL = "https://github.com/pymc-learn/pymc-learn"


classifiers = [
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Operating System :: OS Independent'
]

PROJECT_ROOT = dirname(realpath(__file__))

with open(join(PROJECT_ROOT, 'README.rst'), encoding='utf-8') as r:
    readme = r.read()

REQUIREMENTS_FILE = join(PROJECT_ROOT, 'requirements.txt')

with open(REQUIREMENTS_FILE) as f:
    install_reqs = f.read().splitlines()

if sys.version_info < (3, 4):
    install_reqs.append('enum34')


def get_version():
    VERSIONFILE = join('pmlearn', '__init__.py')
    lines = open(VERSIONFILE, 'rt').readlines()
    version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in lines:
        mo = re.search(version_regex, line, re.M)
        if mo:
            return mo.group(1)
    raise RuntimeError('Unable to find version in %s.' % (VERSIONFILE,))


with open('AUTHORS.txt') as a:
    # reSt-ify the authors list
    authors = ''
    for author in a.read().split('\n'):
        authors += '| '+author+'\n'

with open('LICENSE') as l:
    license = l.read()


if __name__ == "__main__":
    setup(
        name=DISTNAME,
        version=get_version(),
        description=DESCRIPTION,
        long_description=readme,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        license=license,
        packages=find_packages(),
        package_data={'docs': ['*']},
        include_package_data=True,
        zip_safe=False,
        install_requires=install_reqs,
        classifiers=classifiers
        )


"""Kernels for Gaussian process regression and classification.
"""

# Authors: Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import pymc3 as pm


class RBF(pm.gp.cov.ExpQuad):
    """Radial-basis function kernel from ``pymc3.gp.cov.ExpQuad``

    The RBF kernel is a stationary kernel. It is also known as the
    "squared exponential" kernel. It is parameterized by a length-scale
    parameter length_scale>0, which can either be a scalar (isotropic variant
    of the kernel) or a vector with the same number of dimensions as the inputs
    X (anisotropic variant of the kernel).

    The kernel is given by:

    k(x_i, x_j) = exp(-1 / 2 d(x_i / length_scale, x_j / length_scale)^2)

    This kernel is infinitely differentiable, which implies that GPs with this
    kernel as covariance function have mean square derivatives of all orders,
    and are thus very smooth.
    """


class DotProduct(pm.gp.cov.Exponential):
    """Dot-Product kernel from ``pymc3.gp.cov.Exponential``

    The DotProduct kernel is non-stationary and can be obtained from linear
    regression by putting N(0, 1) priors on the coefficients of x_d (d = 1, . .
    . , D) and a prior of N(0, \sigma_0^2) on the bias. The DotProduct kernel
    is invariant to a rotation of the coordinates about the origin, but not
    translations. It is parameterized by a parameter sigma_0^2. For
    sigma_0^2 =0, the kernel is called the homogeneous linear kernel, otherwise
    it is inhomogeneous.

    The kernel is given by
    k(x_i, x_j) = sigma_0 ^ 2 + x_i \cdot x_j

    The DotProduct kernel is commonly combined with exponentiation.
    """


class WhiteKernel(pm.gp.cov.WhiteNoise):
    """White kernel from ``pymc3.gp.cov.WhiteNoise..

    The main use-case of this kernel is as part of a sum-kernel where it
    explains the noise-component of the signal. Tuning its parameter
    corresponds to estimating the noise-level.

    k(x_1, x_2) = noise_level if x_1 == x_2 else 0

    """
    def __init__(self, noise_level=1.0):
        super(WhiteKernel, self).__init__(sigma=noise_level)



"""
Naive Bayes models.
"""

# Authors: Remi Louf <remilouf@gmail.com>, <remi@sounds.am>
#          Daniel Emaasit <daniel.emaasit@gmail.com>
#
# License: BSD 3 clause

import functools as ft

import numpy as np
import pymc3 as pm
import scipy.stats
import theano

from ..exceptions import NotFittedError
from ..base import BayesianModel, BayesianClassifierMixin


class GaussianNBClassifierMixin(BayesianClassifierMixin):
    """Mixin class for naive Bayes classifiers

    """
    def fit(self, X, y, inference_type='advi', minibatch_size=None,
            inference_args=None):
        """
        Train the Naive Bayes model.

        Parameters
        ----------
        X : numpy array, shape [num_training_samples, num_pred].
           Contains the data points.

        y : numpy array, shape [num_training_samples,].
           Contains the category of the data points.

        inference_type : string, specifies which inference method to call.
            Default is 'advi'. Currently, only 'advi' and 'nuts'
            are implemented.

        minibatch_size : int, number of samples to include in each minibatch
            for ADVI. Defaults to None so minibatch is not run by default.

        inference_args : dict, arguments to be passed to the inference methods.
            Check the PyMC3 documentation.

        Returns
        -------
        The current instance of the GaussianNB class.
        """
        self.num_training_samples, self.num_pred = X.shape
        self.num_cats = len(np.unique(y))
        self.inference_type = inference_type

        if not inference_args:
            inference_args = self._set_default_inference_args()

        if not self.cached_model:
            self.cached_model = self.create_model()

        if minibatch_size:
            with self.cached_model:
                minibatches = {
                    self.shared_vars['model_input']: pm.Minibatch(
                        X, batch_size=minibatch_size),
                    self.shared_vars['model_output']: pm.Minibatch(
                        y, batch_size=minibatch_size),
                }

                inference_args['more_replacements'] = minibatches
        else:
            self._set_shared_vars({'model_input': X, 'model_output': y})

        self._inference(inference_type, inference_args)

        return self

    def predict_proba(self, X):
        """
        Predicts the probabilities that data points belong to each category.

        Given a new data point :math:`\\vec{x}`, we want to estimate the
        probability that it belongs to a category :math:`c`.
        Following the notations in [1], the probability
        reads:

        .. math::

            P(y=c|\\vec{x}, \\mathcal{D}) = P(y=c|\\mathcal{D}) \\prod_{j=1}^{n_{dims}} \\
                                            P(x_j|y=c, \\mathcal{D})

        We previously used the data :math:`\\mathcal{D}` to estimate the
        distribution of the parameters :math:`\\vec{\\mu}`, :math:`\\vec{\\pi}`
        and :math:`\\vec{\\sigma}`. To compute the above probability, we need
        to integrate over the values of these parameters:

        .. math::

            P(y=c|\\vec{x}, \\mathcal{D}) = \\left[\\int Cat(y=c|\\vec{\\pi})P(\\vec{\\pi}|\\
                                            \\mathcal{D})\\mathrm{d}\\vec{\\pi}\\right]
                                            \\int P(\\vec{x}|\\vec{\\mu}, \\vec{\\sigma})\\
                                            P(\\vec{\\mu}|\\mathcal{D})\\
                                            P(\\vec{\\sigma}|\\mathcal{D})\\
                                            \\mathrm{d}\\vec{\\mu}\\mathrm{d}\\vec{\\sigma}

        Parameters
        ----------
        X : numpy array, shape [num_training_samples, num_pred].
           Contains the points for which we want to predict the class

        Returns
        -------
        A numpy array of shape [num_training_samples, num_cats] that contains
        the probabilities that each sample belong to each category.

        References
        ----------
        .. [1] Murphy, K. P. (2012). Machine learning: a probabilistic
        perspective.
        """

        if self.trace is None:
            raise NotFittedError("Run fit() on the model before predict()")

        posterior_prediction = np.array([])
        for x in X:
            prob_per_sample = scipy.stats.norm(self.trace['mu'],
                                               self.trace['sigma']).pdf(x)
            prob_per_feature = [
                    np.sum(prob_per_sample[:, :, i], axis=0)/len(
                        self.trace['mu']) for i in range(self.num_pred)
            ]
            prob_per_class = self.normalize(ft.reduce(
                lambda x, y: x*y, prob_per_feature))
            if len(posterior_prediction) == 0:
                posterior_prediction = prob_per_class
            else:
                posterior_prediction = np.vstack((posterior_prediction,
                                                  prob_per_class))

        return posterior_prediction

    def predict(self, X):
        """
        Classify new data with a trained Naive Bayes model. The output is the
        point estimate of the posterior predictive distribution that
        corresponds to the one-hot loss function.

        Parameters
        ----------
        X : numpy array, shape [num_training_samples, num_pred].
           Contains the data to classify.

        Returns
        -------
        A numpy array of shape [num_training_samples,] that contains the
        predicted class to which the data points belong.
        """
        proba = self.predict_proba(X)
        predictions = np.argmax(proba, axis=1)
        return predictions

    def normalize(self, array):
        """
        Normalize values in the array to get probabilities.
        Parameters
        ----------
        array : numpy array of shape [1,]
        Returns
        -------
        A normalized array
        """
        return array / np.sum(array)


class GaussianNB(BayesianModel, GaussianNBClassifierMixin):
    """Gaussian Naive Bayes (GaussianNB) classification built using PyMC3.

    The Gaussian Naive Bayes algorithm assumes that the random variables
    that describe each class and each feature are independent and distributed
    according to Normal distributions.

    Example
    -------
    >>> import numpy as np
    >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    >>> Y = np.array([1, 1, 1, 2, 2, 2])
    >>> from pmlearn.naive_bayes import GaussianNB
    >>> clf = GaussianNB()
    >>> clf.fit(X, Y)
    GaussianNB(priors=None, var_smoothing=1e-09)
    >>> print(clf.predict([[-0.8, -1]]))
    [1]
    >>> clf_pf = GaussianNB()
    >>> clf_pf.partial_fit(X, Y, np.unique(Y))
    GaussianNB(priors=None, var_smoothing=1e-09)
    >>> print(clf_pf.predict([[-0.8, -1]]))
    [1]

    See the documentation of the `create_model` method for details on the model
    itself.
    """

    def __init__(self):
        super(GaussianNB, self).__init__()

    def create_model(self):
        """
        Creates and returns the PyMC3 model.

        We note :math:`x_{jc}` the value of the j-th element of the data
        vector :math:`x` conditioned on x belonging to the class :math:`c`.
        The Gaussian Naive Bayes algorithm models :math:`x_{jc}` as:

        .. math::

            x_{jc} \\sim Normal(\\mu_{jc}, \\sigma_{jc})

        While the probability that :math:`x` belongs to the
        class :math:`c` is given by the categorical distribution:

        .. math::

            P(y=c|x_i) = Cat(\\pi_1, \dots, \\pi_C)

        where :math:`\pi_i` is the probability that a vector belongs
        to category :math:`i`.

        We assume that the :math:`\pi_i` follow a Dirichlet distribution:

        .. math::

            \\pi \\sim Dirichlet(\\alpha)

        with hyperparameter :math:`\\alpha = [1, .., 1]`. The :math:`\\mu_{jc}`
        are sampled from a Normal distribution centred on :math:`0` with
        variance :math:`100`, and the :math:`\\sigma_{jc}` are sampled from a
        HalfNormal distribuion of variance :math:`100`:

        .. math::

            \\mu_{jc} \\sim Normal(0, 100)

            \\sigma_{jc} \\sim HalfNormal(100)

        Note that the Gaussian Naive Bayes model is equivalent to a Gaussian
        mixture with a diagonal covariance [1].

        Returns
        -------
        A PyMC3 model

        References
        ----------
        .. [1] Murphy, K. P. (2012). Machine learning: a probabilistic
        perspective.
        """

        # The data
        X = theano.shared(np.zeros((self.num_training_samples, self.num_pred)))
        y = theano.shared(np.zeros(self.num_training_samples, dtype=int))

        self.shared_vars = {
            'model_input': X,
            'model_output': y
        }

        model = pm.Model()
        with model:
            # Priors
            alpha = np.ones(self.num_cats)
            pi = pm.Dirichlet('pi', alpha, shape=self.num_cats)
            mu = pm.Normal('mu', mu=0, sd=100, shape=(self.num_cats,
                                                      self.num_pred))
            sigma = pm.HalfNormal('sigma', 100, shape=(self.num_cats,
                                                       self.num_pred))

            # Assign classes to data points
            z = pm.Categorical('z', pi, shape=self.num_training_samples,
                               observed=y)

            # The components are independent and normally distributed
            xi = pm.Normal('xi', mu=mu[z], sd=sigma[z], observed=X)

        return model

    def save(self, file_prefix):
        params = {
            'inference_type': self.inference_type,
            'num_cats': self.num_cats,
            'num_pred': self.num_pred,
            'num_training_samples': self.num_training_samples
        }
        super(GaussianNB, self).save(file_prefix, params)

    def load(self, file_profile):
        params = super(GaussianNB, self).load(file_profile,
                                              load_custom_params=True)

        self.inference_type = params['inference_type']
        self.num_cats = params['num_cats']
        self.num_pred = params['num_pred']
        self.num_training_samples = params['num_training_samples']
